@@ -35,25 +35,23 @@ typedef int ssize_t;
 #define sigemptyset(s)
 #define sigfillset(s)
 
-typedef pthread_mutex_t xmutex_t;
+typedef pthread_mutex_t mutex_t;
 #define X_MUTEX_INIT           PTHREAD_MUTEX_INITIALIZER
-#define X_MUTEX_CREATE(mutex)  pthread_mutex_init (&(mutex), 0)
 #define X_LOCK(mutex)          pthread_mutex_lock (&(mutex))
 #define X_UNLOCK(mutex)        pthread_mutex_unlock (&(mutex))
 
-typedef pthread_cond_t xcond_t;
+typedef pthread_cond_t cond_t;
 #define X_COND_INIT                     PTHREAD_COND_INITIALIZER
-#define X_COND_CREATE(cond)		pthread_cond_init (&(cond), 0)
 #define X_COND_SIGNAL(cond)             pthread_cond_signal (&(cond))
 #define X_COND_WAIT(cond,mutex)         pthread_cond_wait (&(cond), &(mutex))
 #define X_COND_TIMEDWAIT(cond,mutex,to) pthread_cond_timedwait (&(cond), &(mutex), &(to))
 
-typedef pthread_t xthread_t;
+typedef pthread_t thread_t;
 #define X_THREAD_PROC(name) void *name (void *thr_arg)
 #define X_THREAD_ATFORK(a,b,c)
 
 static int
-thread_create (xthread_t *tid, void *(*proc)(void *), void *arg)
+thread_create (thread_t *tid, void *(*proc)(void *), void *arg)
 {
   int retval;
   pthread_attr_t attr;
@@ -96,31 +94,22 @@ thread_create (xthread_t *tid, void *(*proc)(void *), void *arg)
 #include <limits.h>
 #include <pthread.h>
 
-typedef pthread_mutex_t xmutex_t;
+typedef pthread_mutex_t mutex_t;
 #if __linux && defined (PTHREAD_ADAPTIVE_MUTEX_INITIALIZER_NP)
-# define X_MUTEX_INIT		PTHREAD_ADAPTIVE_MUTEX_INITIALIZER_NP
-# define X_MUTEX_CREATE(mutex)						\
-  do {									\
-    pthread_mutexattr_t attr;						\
-    pthread_mutexattr_init (&attr);					\
-    pthread_mutexattr_settype (&attr, PTHREAD_MUTEX_ADAPTIVE_NP);	\
-    pthread_mutex_init (&(mutex), &attr);				\
-  } while (0)
+# define X_MUTEX_INIT PTHREAD_ADAPTIVE_MUTEX_INITIALIZER_NP
 #else
-# define X_MUTEX_INIT		PTHREAD_MUTEX_INITIALIZER
-# define X_MUTEX_CREATE(mutex)	pthread_mutex_init (&(mutex), 0)
+# define X_MUTEX_INIT PTHREAD_MUTEX_INITIALIZER
 #endif
-#define X_LOCK(mutex)		pthread_mutex_lock   (&(mutex))
-#define X_UNLOCK(mutex)		pthread_mutex_unlock (&(mutex))
+#define X_LOCK(mutex)   pthread_mutex_lock   (&(mutex))
+#define X_UNLOCK(mutex) pthread_mutex_unlock (&(mutex))
 
-typedef pthread_cond_t xcond_t;
-#define X_COND_INIT			PTHREAD_COND_INITIALIZER
-#define X_COND_CREATE(cond)		pthread_cond_init (&(cond), 0)
-#define X_COND_SIGNAL(cond)		pthread_cond_signal (&(cond))
-#define X_COND_WAIT(cond,mutex)		pthread_cond_wait (&(cond), &(mutex))
-#define X_COND_TIMEDWAIT(cond,mutex,to)	pthread_cond_timedwait (&(cond), &(mutex), &(to))
+typedef pthread_cond_t cond_t;
+#define X_COND_INIT PTHREAD_COND_INITIALIZER
+#define X_COND_SIGNAL(cond) pthread_cond_signal (&(cond))
+#define X_COND_WAIT(cond,mutex) pthread_cond_wait (&(cond), &(mutex))
+#define X_COND_TIMEDWAIT(cond,mutex,to) pthread_cond_timedwait (&(cond), &(mutex), &(to))
 
-typedef pthread_t xthread_t;
+typedef pthread_t thread_t;
 #define X_THREAD_PROC(name) static void *name (void *thr_arg)
 #define X_THREAD_ATFORK(prepare,parent,child) pthread_atfork (prepare, parent, child)
 
@@ -129,12 +118,12 @@ typedef pthread_t xthread_t;
 # define PTHREAD_STACK_MIN 0
 #endif
 
-#ifndef X_STACKSIZE
-# define X_STACKSIZE sizeof (long) * 4096
+#ifndef XTHREAD_STACKSIZE
+# define XTHREAD_STACKSIZE sizeof (long) * 4096
 #endif
 
 static int
-thread_create (xthread_t *tid, void *(*proc)(void *), void *arg)
+thread_create (thread_t *tid, void *(*proc)(void *), void *arg)
 {
   int retval;
   sigset_t fullsigset, oldsigset;
@@ -142,7 +131,11 @@ thread_create (xthread_t *tid, void *(*proc)(void *), void *arg)
 
   pthread_attr_init (&attr);
   pthread_attr_setdetachstate (&attr, PTHREAD_CREATE_DETACHED);
-  pthread_attr_setstacksize (&attr, PTHREAD_STACK_MIN < X_STACKSIZE ? X_STACKSIZE : PTHREAD_STACK_MIN);
+
+  if (XTHREAD_STACKSIZE > 0)
+    pthread_attr_setstacksize (&attr, PTHREAD_STACK_MIN > (XTHREAD_STACKSIZE)
+                                      ? PTHREAD_STACK_MIN : (XTHREAD_STACKSIZE));
+
 #ifdef PTHREAD_SCOPE_PROCESS
   pthread_attr_setscope (&attr, PTHREAD_SCOPE_PROCESS);
 #endif
